import abc


__version__= "$Id: $"



class AbstractGeneralCommandSet(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def connectTCPIP(self, hostname, port=50000):
        assert False


    @abc.abstractmethod
    def closeConnection(self):
        assert False


    @abc.abstractmethod
    def gcsCommand(self, commandAsString):
        assert False


    @abc.abstractmethod
    def getVersion(self):
        assert False


    @abc.abstractmethod
    def getAxesIdentifiers(self):
        assert False


    @abc.abstractmethod
    def getNumberOfInputSignalChannels(self):
        assert False


    @abc.abstractmethod
    def getNumberOfOutputSignalChannels(self):
        assert False


    @abc.abstractmethod
    def getServoControlMode(self, axesString):
        assert False


    @abc.abstractmethod
    def setServoControlMode(self, axesString, controlMode):
        assert False


    @abc.abstractmethod
    def getControlMode(self, channels):
        assert False


    @abc.abstractmethod
    def setControlMode(self, channels, controlMode):
        assert False


    @abc.abstractmethod
    def enableControlMode(self, channels):
        assert False


    @abc.abstractmethod
    def disableControlMode(self, channels):
        assert False


    @abc.abstractmethod
    def echo(self, message):
        assert False


    @abc.abstractmethod
    def getLowerVoltageLimit(self, channels):
        assert False


    @abc.abstractmethod
    def setLowerVoltageLimit(self, channels, lowerVoltage):
        assert False


    @abc.abstractmethod
    def getUpperVoltageLimit(self, channels):
        assert False


    @abc.abstractmethod
    def setUpperVoltageLimit(self, channels, upperVoltage):
        assert False


    @abc.abstractmethod
    def getPosition(self, axesString):
        assert False


    @abc.abstractmethod
    def getVoltages(self, channels):
        assert False


    @abc.abstractmethod
    def getOpenLoopAxisValue(self, axesString):
        assert False


    @abc.abstractmethod
    def setOpenLoopAxisValue(self, axesString, amplitudeInVolt):
        assert False


    @abc.abstractmethod
    def setRelativeOpenLoopAxisValue(self, axesString, offsetInVolt):
        assert False


    @abc.abstractmethod
    def getTargetPosition(self, axesString):
        assert False


    @abc.abstractmethod
    def setTargetPosition(self, axesString, position):
        assert False


    @abc.abstractmethod
    def setTargetRelativeToCurrentPosition(self, axesString, offset):
        assert False


    @abc.abstractmethod
    def getVolatileMemoryParameters(self, itemId, parameterId):
        assert False


    @abc.abstractmethod
    def getAllDataRecorderOptions(self):
        assert False


    @abc.abstractmethod
    def getNumberOfRecorderTables(self):
        assert False


    @abc.abstractmethod
    def setDataRecorderConfiguration(self, dataRecorderConfiguration):
        assert False


    @abc.abstractmethod
    def getDataRecorderConfiguration(self):
        assert False


    @abc.abstractmethod
    def getRecordedDataValues(self, howManyPoints, startFromPoint=1):
        assert False


    @abc.abstractmethod
    def startRecordingInSyncWithWaveGenerator(self):
        assert False


    @abc.abstractmethod
    def getNumberOfWaveGenerators(self):
        assert False


    @abc.abstractmethod
    def getWaveGeneratorStartStopMode(self):
        assert False


    @abc.abstractmethod
    def setWaveGeneratorStartStopMode(self, startModeArray):
        assert False


    @abc.abstractmethod
    def clearWaveTableData(self, waveTableIdsArray):
        assert False


    @abc.abstractmethod
    def getConnectionOfWaveTableToWaveGenerator(self, waveGeneratorsArray):
        assert False


    @abc.abstractmethod
    def setConnectionOfWaveTableToWaveGenerator(self,
                                                waveGeneratorsArray,
                                                waveTableIdsArray):
        assert False

    @abc.abstractmethod
    def setSinusoidalWaveform(self,
                              waveTableId,
                              append,
                              lengthInPoints,
                              amplitudeOfTheSineCurve,
                              offsetOfTheSineCurve,
                              wavelengthOfTheSineCurveInPoints,
                              startPoint,
                              curveCenterPoint):
        assert False



    @abc.abstractmethod
    def setRecordTableRate(self, recordTableRateInServoLoopCycles=1):
        assert False


    @abc.abstractmethod
    def getRecordTableRate(self):
        assert False


    @abc.abstractmethod
    def getServoUpdateTimeInSeconds(self):
        assert False



    @abc.abstractmethod
    def setWaveGeneratorTableRate(self,
                                  waveGeneratorTableRateInServoLoopCycles):
        assert False


    @abc.abstractmethod
    def getWaveGeneratorTableRate(self):
        assert False


    @abc.abstractmethod
    def getOverflowState(self):
        assert False


    @abc.abstractmethod
    def setDataRecoderTriggerSource(self, source, value):
        assert False


    @abc.abstractmethod
    def getDataRecorderTriggerSource(self):
        assert False


    @abc.abstractmethod
    def startStepAndResponseMeasurement(self, axisString, amplitude):
        assert False


    @abc.abstractmethod
    def startImpulseAndResponseMeasurement(self, axisString, amplitude):
        assert False





from pi_gcs.gcs2 import GeneralCommandSet2


__version__= "$Id: $"




class PhysikInstrumenteE518Client(object):

    def __init__(self, hostname, generalCommandSet):
        self._hostname= hostname
        self._generalCommandSet= generalCommandSet


    def foo(self):
        return 'foo'


    def connect(self):
        self.connectTCPIP(self._hostname)





__version__= "$Id: $"


class RecordOption(object):
    UNKNOWN_OPTION= 0
    TARGET_POSITION_OF_AXIS= 1
    REAL_POSITION_OF_AXIS= 2
    POSITION_ERROR_OF_AXIS= 3
    VOLTAGE_OF_PIEZO_CHANNEL= 7
    CONTROL_OUTPUT_OF_AXIS= 15
    DAC_OF_PIEZO_CHANNEL= 16
    SENSOR_VALUE= 17
    SENSOR_FILTERED= 18
    TARGET_POSITION_SLEW_RATE_LIMITED= 22
    DRIFT_COMPENSATION_OFFSET= 32
    ADC_OF_ANALOG_INPUT= 81


    ALL_OPTIONS= [UNKNOWN_OPTION,
                  TARGET_POSITION_OF_AXIS,
                  REAL_POSITION_OF_AXIS,
                  POSITION_ERROR_OF_AXIS,
                  VOLTAGE_OF_PIEZO_CHANNEL,
                  CONTROL_OUTPUT_OF_AXIS,
                  DAC_OF_PIEZO_CHANNEL,
                  SENSOR_VALUE,
                  SENSOR_FILTERED,
                  TARGET_POSITION_SLEW_RATE_LIMITED,
                  DRIFT_COMPENSATION_OFFSET,
                  ADC_OF_ANALOG_INPUT]


class DataRecorderConfiguration(object):



    def __init__(self):
        self._table= {}


    def setTable(self, recordTable, source, recordOption):
        assert recordOption in RecordOption.ALL_OPTIONS,\
            'unknown record option %d. Valid values %s' % (
                recordOption, str(RecordOption.ALL_OPTIONS))

        self._table[recordTable]= (source, recordOption)


    def getRecordSource(self, recordTable):
        return self._table[recordTable][0]


    def getRecordOption(self, recordTable):
        return self._table[recordTable][1]


    def getTableIds(self):
        return list(self._table.keys())



import numpy as np
from pi_gcs.gcs2 import WaveformGenerator, PIException
from pi_gcs.data_recorder_configuration import DataRecorderConfiguration,\
    RecordOption

__version__= "$Id: $"


class TipTilt2Axis(object):

    AXIS_A= "A"
    AXIS_B= "B"
    ALL_AXES= "A B"
    ALL_CHANNELS= [1, 2, 3]

    def __init__(self, piController, tipTiltConfiguration):
        self._ctrl= piController
        self._cfg= tipTiltConfiguration

        self._origTargetPosition= None
        self._modulationEnabled= False



    def setUp(self):
        self._connectController()
        self._checkNumberOfChannels()
        self._enableRemoteControlMode()
        self._stopWaveformGenerators()
        self._setVoltageLimits()
        self.disableControlLoop()
        self._configure3rdAxisAsPivot()
        self.enableControlLoop()


    def _connectController(self):
        self._ctrl.connectTCPIP(self._cfg.hostname)


    def _checkNumberOfChannels(self):
        assert 3 == self._ctrl.getNumberOfInputSignalChannels()
        assert 3 == self._ctrl.getNumberOfOutputSignalChannels()


    def _enableRemoteControlMode(self):
        self._ctrl.enableControlMode(self.ALL_CHANNELS)


    def _stopWaveformGenerators(self):
        self._ctrl.setWaveGeneratorStartStopMode([0, 0, 0])


    def _setVoltageLimits(self):
        self._ctrl.setLowerVoltageLimit(
            self.ALL_CHANNELS, self._cfg.lowerVoltageLimit)
        self._ctrl.setUpperVoltageLimit(
            self.ALL_CHANNELS, self._cfg.upperVoltageLimit)


    def _configure3rdAxisAsPivot(self):
        pivotAxis= self._ctrl.getAxesIdentifiers()[2]
        self._ctrl.setServoControlMode(pivotAxis, [False])
        self._ctrl.setOpenLoopAxisValue(pivotAxis, self._cfg.pivotValue)


    def enableControlLoop(self):
        self._ctrl.setServoControlMode(self.ALL_AXES, [True, True])


    def disableControlLoop(self):
        self._ctrl.setServoControlMode(self.ALL_AXES, [False, False])


    def isControlLoopEnabled(self):
        return np.all(self._ctrl.getServoControlMode(self.ALL_AXES))


    def _positionConversionCoefficients(self, axisName):
        if axisName == self.AXIS_A:
            offset= self._cfg.positionToMilliRadAxisAOffsetCoeff
            linear= self._cfg.positionToMilliRadAxisALinearCoeff
        elif axisName == self.AXIS_B:
            offset= self._cfg.positionToMilliRadAxisBOffsetCoeff
            linear= self._cfg.positionToMilliRadAxisBLinearCoeff
        else:
            raise KeyError("unknown axis %s")
        return offset, linear


    def _gcsUnitsToMilliRadOneAxis(self, posInGcsUnits, axisName):
        offset, linear= self._positionConversionCoefficients(axisName)
        return linear * posInGcsUnits + offset


    def _gcsUnitsToMilliRad(self, positionInGcsUnits):
        mradA= self._gcsUnitsToMilliRadOneAxis(positionInGcsUnits[0],
                                               self.AXIS_A)
        mradB= self._gcsUnitsToMilliRadOneAxis(positionInGcsUnits[1],
                                               self.AXIS_B)
        return np.array([mradA, mradB])


    def _milliRadToGcsUnitsOneAxis(self, posInMilliRad, axisName):
        offset, linear= self._positionConversionCoefficients(axisName)
        return (posInMilliRad - offset) / linear


    def _milliRadToGcsUnits(self, positionInMilliRad):
        posA= self._milliRadToGcsUnitsOneAxis(positionInMilliRad[0],
                                              self.AXIS_A)
        posB= self._milliRadToGcsUnitsOneAxis(positionInMilliRad[1],
                                              self.AXIS_B)
        return np.array([posA, posB])


    def getPosition(self):
        return self._gcsUnitsToMilliRad(self._ctrl.getPosition(self.ALL_AXES))


    def getTargetPosition(self):
        return self._gcsUnitsToMilliRad(
            self._ctrl.getTargetPosition(self.ALL_AXES))


    def setTargetPosition(self, positionInMilliRad):
        return self._ctrl.setTargetPosition(
            self.ALL_AXES,
            self._milliRadToGcsUnits(positionInMilliRad))


    def getVoltages(self):
        return self._ctrl.getVoltages(self.ALL_CHANNELS)


    def startSinusoidalModulation(self,
                                  radiusInMilliRad,
                                  frequencyInHz,
                                  phasesInRadians,
                                  centerInMilliRad):
        self._origTargetPosition= centerInMilliRad
        self.stopModulation()

        assert np.ptp(self._ctrl.getWaveGeneratorTableRate()) == 0, \
            "wave generator table rate must be the same for every table"
        wgtr= self._ctrl.getWaveGeneratorTableRate()[0]
        timestep= self._ctrl.getServoUpdateTimeInSeconds() * wgtr

#         periodInSec= 1./ frequencyInHz
#         lengthInPoints= periodInSec/ timestep
#         peakOfTheSineCurve= self._milliRadToGcsUnits(
#             self.getTargetPosition() + radiusInMilliRad)
#         offsetOfTheSineCurve= self._milliRadToGcsUnits(
#             self.getTargetPosition() - radiusInMilliRad)
#         amplitudeOfTheSineCurve= peakOfTheSineCurve - offsetOfTheSineCurve
#         wavelengthOfTheSineCurveInPoints= periodInSec/ timestep
#         startPoint= np.array([0, 0.25])* wavelengthOfTheSineCurveInPoints
#         curveCenterPoint= 0.5* wavelengthOfTheSineCurveInPoints

        self._ctrl.clearWaveTableData([1, 2, 3])
        self._setSinusoidalWaveform(1, timestep, radiusInMilliRad[0],
                                    frequencyInHz, phasesInRadians[0],
                                    centerInMilliRad[0])
        self._setSinusoidalWaveform(2, timestep, radiusInMilliRad[1],
                                    frequencyInHz, phasesInRadians[1],
                                    centerInMilliRad[1])
#         self._ctrl.setSinusoidalWaveform(
#             1, WaveformGenerator.CLEAR, lengthInPoints,
#             amplitudeOfTheSineCurve[0], offsetOfTheSineCurve[0],
#             wavelengthOfTheSineCurveInPoints, startPoint[0], curveCenterPoint)
#         self._ctrl.setSinusoidalWaveform(
#             2, WaveformGenerator.CLEAR, lengthInPoints,
#             amplitudeOfTheSineCurve[1], offsetOfTheSineCurve[1],
#             wavelengthOfTheSineCurveInPoints, startPoint[1], curveCenterPoint)
        self._ctrl.setConnectionOfWaveTableToWaveGenerator([1, 2], [1, 2])
        self._ctrl.setWaveGeneratorStartStopMode([1, 1, 0])
        self._modulationEnabled= True


    def _getAxisFromWaveTableId(self, waveTableId):
        if waveTableId == 1:
            return self.AXIS_A
        elif waveTableId == 2:
            return self.AXIS_B
        else:
            raise ValueError("Unknown waveTableId %d" % waveTableId)


    def _setSinusoidalWaveform(self, waveTableId, timeStepInSec,
                               amplitudeInMilliRad, frequencyInHz,
                               phaseInRadians, offsetInMilliRad):
        axisName= self._getAxisFromWaveTableId(waveTableId)
        periodInSec= 1./ frequencyInHz
        wgtr= self._ctrl.getWaveGeneratorTableRate()[0]
        timestep= self._ctrl.getServoUpdateTimeInSeconds() * wgtr

        lengthInPoints= periodInSec/ timestep
        peakOfTheSineCurve= self._milliRadToGcsUnitsOneAxis(
            offsetInMilliRad + amplitudeInMilliRad, axisName)
        valleyOfTheSineCurve= self._milliRadToGcsUnitsOneAxis(
            offsetInMilliRad - amplitudeInMilliRad, axisName)
        amplitudeOfTheSineCurve= peakOfTheSineCurve - valleyOfTheSineCurve
        wavelengthOfTheSineCurveInPoints= periodInSec/ timestep
        startPoint= phaseInRadians/ (2* np.pi) * \
            wavelengthOfTheSineCurveInPoints
        curveCenterPoint= 0.5* wavelengthOfTheSineCurveInPoints
        self._ctrl.setSinusoidalWaveform(
            waveTableId, WaveformGenerator.CLEAR, lengthInPoints,
            amplitudeOfTheSineCurve, valleyOfTheSineCurve,
            wavelengthOfTheSineCurveInPoints, startPoint, curveCenterPoint)




    def stopModulation(self):
        self._stopWaveformGenerators()
        self._modulationEnabled= False
        if self._origTargetPosition is not None:
            self.setTargetPosition(self._origTargetPosition)


    def isModulationEnabled(self):
        return self._modulationEnabled


    def _defaultDataRecorderConfiguration(self):
        dataRecorderCfg= DataRecorderConfiguration()
        dataRecorderCfg.setTable(1, "A", RecordOption.REAL_POSITION_OF_AXIS)
        dataRecorderCfg.setTable(2, "B", RecordOption.REAL_POSITION_OF_AXIS)
        dataRecorderCfg.setTable(3, "A", RecordOption.POSITION_ERROR_OF_AXIS)
        dataRecorderCfg.setTable(4, "B", RecordOption.POSITION_ERROR_OF_AXIS)
        dataRecorderCfg.setTable(5, "A", RecordOption.TARGET_POSITION_OF_AXIS)
        dataRecorderCfg.setTable(6, "B", RecordOption.TARGET_POSITION_OF_AXIS)
        return dataRecorderCfg


    def _configureDataRecoders(self, dataRecorderCfg=None):
        if dataRecorderCfg is None:
            dataRecorderCfg= self._defaultDataRecorderConfiguration()
        self._ctrl.setDataRecorderConfiguration(dataRecorderCfg)


    def getDataRecorderConfiguration(self):
        return self._ctrl.getDataRecorderConfiguration()


    def _convertRecordedDataToMilliRad(self, recData):
        ret= recData.copy()
        recordOptionsToConvert= [RecordOption.TARGET_POSITION_OF_AXIS,
                                 RecordOption.POSITION_ERROR_OF_AXIS,
                                 RecordOption.REAL_POSITION_OF_AXIS]
        cfg= self.getDataRecorderConfiguration()
        ids= cfg.getTableIds()
        for i in np.arange(len(ids)):
            source= cfg.getRecordSource(ids[i])
            option= cfg.getRecordOption(ids[i])
            if option in recordOptionsToConvert:
                ret[i]= self._gcsUnitsToMilliRadOneAxis(ret[i], source)
        return ret


    def getRecordedData(self, howManyPoints, dataRecorderCfg=None):
        self._startDataRecorder(dataRecorderCfg)
        return self._retrieveRecordedData(howManyPoints)


    def _startDataRecorder(self, dataRecorderCfg=None):
        self._configureDataRecoders(dataRecorderCfg)
        self._ctrl.startRecordingInSyncWithWaveGenerator()


    def _retrieveRecordedData(self, howManyPoints):
        timestep= self._ctrl.getServoUpdateTimeInSeconds()
        rtr= self._ctrl.getRecordTableRate()
        timeValues= np.arange(howManyPoints) * timestep * rtr
        recDataGcs= self._ctrl.getRecordedDataValues(howManyPoints, 1)
        recDataMilliRad= self._convertRecordedDataToMilliRad(recDataGcs)
        return np.vstack((timeValues, recDataMilliRad))


    def status(self):
        status={}
        status['POSITION']= self.getPosition()
        status['TARGET']= self.getTargetPosition()
        status['OUTPUT_VOLTAGE']= self.getVoltages()
        status['CONTROL_LOOP_CLOSED']= self.isControlLoopEnabled()
        status['OVERFLOW']= self._ctrl.getOverflowState(self.ALL_AXES)
        return status



import numpy as np
from pi_gcs.abstract_gcs2 import AbstractGeneralCommandSet
from pi_gcs.gcs2 import WaveformGenerator
from pi_gcs.data_recorder_configuration import DataRecorderConfiguration,\
    RecordOption


__version__= "$Id: $"


class FakeGeneralCommandSet(AbstractGeneralCommandSet):

    def __init__(self):
        self._channelsInCloseLoop= {}
        self._lowerVoltageLimits= {}
        self._upperVoltageLimits= {}
        self._numberOfInputSignalChannels= 3
        self._numberOfOutputSignalChannels= 3
        self._axesIdentifiers= ['A', 'B', 'C']
        self._openLoopAxisValue= {}
        self._targetPosition= {}
        self._position= {}
        self._waveGeneratorStartStopMode= {}
        self._rtr= 1
        self._wtr= np.ones(3)
        self.triggerStartRecordingInSyncWithWaveGenerator= 0
        self._waveform= {}
        self._dataRecorderConfig= self._defaultDataRecorderConfiguration()


    def _dictToArray(self, dicto, keys):
        ret= []
        for k in keys:
            ret.append(dicto[k])
        return np.array(ret)


    def _arrayToDict(self, dicto, keys, values):
        valueArray= np.atleast_1d(values)
        assert len(keys) == len(valueArray), \
            "%d %d" % (len(keys), len(valueArray))
        for i in range(len(keys)):
            dicto[keys[i]]= valueArray[i]


    def _axesString2Array(self, axesString):
        return [x.strip() for x in axesString.split(' ')]


    def connectTCPIP(self, hostname, port=50000):
        pass



    def closeConnection(self):
        pass



    def gcsCommand(self, commandAsString):
        pass



    def getVersion(self):
        pass



    def getAxesIdentifiers(self):
        return self._axesIdentifiers


    def getNumberOfInputSignalChannels(self):
        return self._numberOfInputSignalChannels


    def getNumberOfOutputSignalChannels(self):
        return self._numberOfOutputSignalChannels


    def getServoControlMode(self, axesString):
        axes= self._axesString2Array(axesString)
        return self._dictToArray(self._channelsInCloseLoop, axes)


    def setServoControlMode(self, axesString, controlMode):
        axes= self._axesString2Array(axesString)
        self._arrayToDict(self._channelsInCloseLoop, axes, controlMode)


    def getControlMode(self, channels):
        pass



    def setControlMode(self, channels, controlMode):
        pass



    def enableControlMode(self, channels):
        pass



    def disableControlMode(self, channels):
        pass



    def echo(self, message):
        pass



    def getLowerVoltageLimit(self, channels):
        return self._dictToArray(self._lowerVoltageLimits, channels)


    def setLowerVoltageLimit(self, channels, lowerVoltage):
        self._arrayToDict(self._lowerVoltageLimits, channels, lowerVoltage)


    def getUpperVoltageLimit(self, channels):
        return self._dictToArray(self._upperVoltageLimits, channels)


    def setUpperVoltageLimit(self, channels, upperVoltage):
        self._arrayToDict(self._upperVoltageLimits, channels, upperVoltage)


    def getPosition(self, axesString):
        return self.getTargetPosition(axesString)


    def getVoltages(self, channels):
        pass


    def getOpenLoopAxisValue(self, axesString):
        axes= [x.strip() for x in axesString.split(' ')]
        return self._dictToArray(self._openLoopAxisValue, axes)


    def setOpenLoopAxisValue(self, axesString, amplitudeInVolt):
        axes= [x.strip() for x in axesString.split(' ')]
        self._arrayToDict(self._openLoopAxisValue, axes, amplitudeInVolt)


    def setRelativeOpenLoopAxisValue(self, axesString, offsetInVolt):
        pass


    def getTargetPosition(self, axesString):
        axes= [x.strip() for x in axesString.split(' ')]
        return self._dictToArray(self._targetPosition, axes)


    def setTargetPosition(self, axesString, position):
        axes= [x.strip() for x in axesString.split(' ')]
        self._arrayToDict(self._targetPosition, axes, position)


    def setTargetRelativeToCurrentPosition(self, axesString, offset):
        pass


    def getVolatileMemoryParameters(self, itemId, parameterId):
        pass


    def getAllDataRecorderOptions(self):
        pass



    def getNumberOfRecorderTables(self):
        return 11


    def setDataRecorderConfiguration(self, dataRecorderConfiguration):
        self._dataRecorderConfig= dataRecorderConfiguration


    def getDataRecorderConfiguration(self):
        return self._dataRecorderConfig


    def _defaultDataRecorderConfiguration(self):
        dataRecorderCfg= DataRecorderConfiguration()
        dataRecorderCfg.setTable(1, "A", RecordOption.REAL_POSITION_OF_AXIS)
        dataRecorderCfg.setTable(2, "B", RecordOption.REAL_POSITION_OF_AXIS)
        dataRecorderCfg.setTable(3, "A", RecordOption.POSITION_ERROR_OF_AXIS)
        dataRecorderCfg.setTable(4, "B", RecordOption.POSITION_ERROR_OF_AXIS)
        dataRecorderCfg.setTable(5, "A", RecordOption.TARGET_POSITION_OF_AXIS)
        dataRecorderCfg.setTable(6, "B", RecordOption.TARGET_POSITION_OF_AXIS)
        return dataRecorderCfg


    def getRecordedDataValues(self, howManyPoints, startFromPoint=1):
        nRecorders= self.getNumberOfRecorderTables()
        return np.arange(howManyPoints * nRecorders).\
            reshape((nRecorders, howManyPoints))


    def startRecordingInSyncWithWaveGenerator(self):
        self.triggerStartRecordingInSyncWithWaveGenerator+= 1


    def getNumberOfWaveGenerators(self):
        pass


    def getWaveGeneratorStartStopMode(self):
        return self._dictToArray(self._waveGeneratorStartStopMode,
                                 [1, 2, 3])


    def setWaveGeneratorStartStopMode(self, startModeArray):
        self._arrayToDict(self._waveGeneratorStartStopMode,
                          [1, 2, 3],
                          startModeArray)


    def clearWaveTableData(self, waveTableIdsArray):
        pass


    def getConnectionOfWaveTableToWaveGenerator(self, waveGeneratorsArray):
        pass


    def setConnectionOfWaveTableToWaveGenerator(self,
                                                waveGeneratorsArray,
                                                waveTableIdsArray):
        pass


    def setSinusoidalWaveform(self,
                              waveTableId,
                              append,
                              lengthInPoints,
                              amplitudeOfTheSineCurve,
                              offsetOfTheSineCurve,
                              wavelengthOfTheSineCurveInPoints,
                              startPoint,
                              curveCenterPoint):
        '''
        See description of PI_WAV_SIN_P in PI GCS 2.0 DLL doc
        '''
        curveCenterPoint= int(round(curveCenterPoint))
        wavelengthOfTheSineCurveInPoints= \
            int(round(wavelengthOfTheSineCurveInPoints))
        startPoint= int(round(startPoint))
        lengthInPoints= int(round(lengthInPoints))
        assert append == WaveformGenerator.CLEAR, 'only CLEAR implemented'
        assert startPoint >= 0
        assert startPoint < lengthInPoints
        assert curveCenterPoint >= 0
        assert startPoint + curveCenterPoint < lengthInPoints, \
            'startPoint + curveCenterPoint >= lenghtInPoints (%d+%d>=%d)' % (
                startPoint, curveCenterPoint, lengthInPoints)

        ccUp= 0.5* curveCenterPoint
        rampUp= 0.5 * amplitudeOfTheSineCurve* (1 + np.sin(
            np.arange(-ccUp, ccUp) / ccUp * np.pi / 2))
        ccDown= 0.5* (wavelengthOfTheSineCurveInPoints - curveCenterPoint)
        rampDown= 0.5 * amplitudeOfTheSineCurve* (1 - np.sin(
            np.arange(-ccDown, ccDown) / ccDown * np.pi / 2))
        waveform= np.zeros(lengthInPoints) + offsetOfTheSineCurve
        waveform[0: curveCenterPoint]= offsetOfTheSineCurve + rampUp
        waveform[curveCenterPoint: wavelengthOfTheSineCurveInPoints]= \
            offsetOfTheSineCurve + rampDown
        waveform= np.roll(waveform, startPoint)
        self._waveform[waveTableId]= waveform


    def getWaveform(self, waveTableId):
        return self._waveform[waveTableId]


    def setRecordTableRate(self, recordTableRateInServoLoopCycles=1):
        self._rtr= recordTableRateInServoLoopCycles


    def getRecordTableRate(self):
        return self._rtr


    def getServoUpdateTimeInSeconds(self):
        return 40e-6


    def setWaveGeneratorTableRate(self,
                                  waveGeneratorTableRateInServoLoopCycles):
        self._wtr= np.ones(3) * waveGeneratorTableRateInServoLoopCycles


    def getWaveGeneratorTableRate(self):
        return self._wtr


    def getOverflowState(self):
        pass


    def setDataRecoderTriggerSource(self, source, value):
        pass


    def getDataRecorderTriggerSource(self):
        pass


    def startStepAndResponseMeasurement(self, axisString, amplitude):
        pass


    def startImpulseAndResponseMeasurement(self, axisString, amplitude):
        pass





__version__= "$Id: $"


class TipTiltConfiguration():

    HOSTNAME= 'hostname'
    LOWER_VOLTAGE_LIMIT= 'lowerVoltageLimit'
    PIVOT_VALUE= 'pivotValue'
    UPPER_VOLTAGE_LIMIT= 'upperVoltageLimit'
    POSITION_TO_MILLIRAD_AXIS_A_LINEAR_COEFF= \
        'positionToMilliRadAxisALinearCoeff'
    POSITION_TO_MILLIRAD_AXIS_A_OFFSET_COEFF= \
        'positionToMilliRadAxisAOffsetCoeff'
    POSITION_TO_MILLIRAD_AXIS_B_LINEAR_COEFF= \
        'positionToMilliRadAxisBLinearCoeff'
    POSITION_TO_MILLIRAD_AXIS_B_OFFSET_COEFF= \
        'positionToMilliRadAxisBOffsetCoeff'

    def __init__(self):
        self._calib= {}
        self._calib[self.LOWER_VOLTAGE_LIMIT]= [0, 0, 0]
        self._calib[self.UPPER_VOLTAGE_LIMIT]= [100, 100, 100]
        self._calib[self.PIVOT_VALUE]= 100
        self._calib[self.HOSTNAME]= None
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_A_LINEAR_COEFF]= 1.0
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_A_OFFSET_COEFF]= 0.0
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_B_LINEAR_COEFF]= 1.0
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_B_OFFSET_COEFF]= 0.0


    def calibration(self):
        return self._calib


    @property
    def lowerVoltageLimit(self):
        return self._calib[self.LOWER_VOLTAGE_LIMIT]


    @property
    def upperVoltageLimit(self):
        return self._calib[self.UPPER_VOLTAGE_LIMIT]


    @property
    def pivotValue(self):
        return self._calib[self.PIVOT_VALUE]


    @property
    def hostname(self):
        return self._calib[self.HOSTNAME]


    @hostname.setter
    def hostname(self, value):
        self._calib[self.HOSTNAME]= value


    @property
    def positionToMilliRadAxisALinearCoeff(self):
        return self._calib[self.POSITION_TO_MILLIRAD_AXIS_A_LINEAR_COEFF]


    @positionToMilliRadAxisALinearCoeff.setter
    def positionToMilliRadAxisALinearCoeff(self, value):
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_A_LINEAR_COEFF]= value


    @property
    def positionToMilliRadAxisAOffsetCoeff(self):
        return self._calib[self.POSITION_TO_MILLIRAD_AXIS_A_OFFSET_COEFF]


    @positionToMilliRadAxisAOffsetCoeff.setter
    def positionToMilliRadAxisAOffsetCoeff(self, value):
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_A_OFFSET_COEFF]= value


    @property
    def positionToMilliRadAxisBLinearCoeff(self):
        return self._calib[self.POSITION_TO_MILLIRAD_AXIS_B_LINEAR_COEFF]


    @positionToMilliRadAxisBLinearCoeff.setter
    def positionToMilliRadAxisBLinearCoeff(self, value):
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_B_LINEAR_COEFF]= value


    @property
    def positionToMilliRadAxisBOffsetCoeff(self):
        return self._calib[self.POSITION_TO_MILLIRAD_AXIS_B_OFFSET_COEFF]


    @positionToMilliRadAxisBOffsetCoeff.setter
    def positionToMilliRadAxisBOffsetCoeff(self, value):
        self._calib[self.POSITION_TO_MILLIRAD_AXIS_B_OFFSET_COEFF]= value






from ctypes.util import find_library
import ctypes
from ctypes import c_int, c_bool, c_char, c_char_p, c_double, c_uint
import numpy as np
from pi_gcs.data_recorder_configuration import DataRecorderConfiguration
from pi_gcs.abstract_gcs2 import AbstractGeneralCommandSet


__version__= "$Id: $"

CHANNEL_OFFLINE= 0
CHANNEL_ONLINE= 1


class PIException(Exception):
    pass


class PIConnectionError(PIException):
    pass


class CTypeArray():

    def __init__(self, ctype, array):
        assert isinstance(array, (np.ndarray, list, tuple))
        self._ctype= ctype
        self._ret= (self._ctype * len(array))()
        for i in range(len(array)):
            self._ret[i]= array[i]


    def from_param(self):
        return self._ret


    def toNumpyArray(self):
        return np.array([x for x in self._ret])


class CIntArray(CTypeArray):
    def __init__(self, array):
        CTypeArray.__init__(self, c_int, np.array(array, dtype=np.int))


class CUnsignedIntArray(CTypeArray):
    def __init__(self, array):
        CTypeArray.__init__(self, c_uint, array)


class CDoubleArray(CTypeArray):
    def __init__(self, array):
        CTypeArray.__init__(self, c_double, array)


class CBoolArray(CTypeArray):
    def __init__(self, array):
        CTypeArray.__init__(self, c_bool, array)


class CCharArray(CTypeArray):
    def __init__(self, array):
        CTypeArray.__init__(self, c_char, array)



class Channels(CIntArray):
    def __init__(self, channels):
        CIntArray.__init__(self, channels)



class WaveformGenerator(object):
    CLEAR= 0
    APPEND= 1
    ADD= 2

    ALL= [CLEAR, APPEND, ADD]


class GeneralCommandSet2(AbstractGeneralCommandSet):

    GCS_TRUE= 1
    GCS_FALSE= 0

    def __init__(self):
        self._hostname= None
        self._port= None
        self._lib= None
        self._id= None
        self._axes= ctypes.c_char_p(b"A B C")
        self._channels= (ctypes.c_int * 3)(1, 2, 3)

        self._importLibrary()


    def _importLibrary(self):
        piLibName= 'pi_pi_gcs2'
        pilib= find_library(piLibName)
        if pilib is None:
            raise PIException("Library %s not found" % piLibName)
        self._lib= ctypes.CDLL(pilib)


    def _errAsString(self, errorCode):
        bufSize= 256
        s=ctypes.create_string_buffer(b'\000' * bufSize)

        ret= self._lib.PI_TranslateError(
            errorCode, s, bufSize)
        if ret != 1:
            return "Unknown error (%d)" % errorCode
        return "%s (%d)" % (s.value.decode(), errorCode)


    def _convertErrorToException(self, returnValue, expectedReturn=GCS_TRUE):
        if returnValue != expectedReturn:
            errorId= self._lib.PI_GetError(self._id)
            if errorId != 0:
                errMsg= ""
                while errorId != 0:
                    errMsg+= self._errAsString(errorId)+ " "
                    errorId= self._lib.PI_GetError(self._id)
                raise PIException("%s" % errMsg)


    def _convertCArrayToNumpyArray(self, cArray):
        return np.array([x for x in cArray.array()])


    def _toBool(self, value):
        assert value in [0, 1]
        return True if value == 1 else False


    def _getterChannels(self, channels, gcsFunction, valueArrayClass):
        chArray= np.atleast_1d(channels)
        value= valueArrayClass([0] * len(chArray))
        gcsFunction.argtypes= [c_int, CIntArray, valueArrayClass, c_int]
        self._convertErrorToException(
            gcsFunction(self._id,
                        CIntArray(chArray),
                        value,
                        len(chArray)))
        return value.toNumpyArray()


    def _setterChannels(self, channels, value, gcsFunction, valueArrayClass):
        valueArray= np.atleast_1d(value)
        assert len(channels) == len(valueArray)
        gcsFunction.argtypes= [c_int, CIntArray, valueArrayClass, c_int]
        self._convertErrorToException(
            gcsFunction(self._id,
                        CIntArray(channels),
                        valueArrayClass(valueArray),
                        len(channels)))


    def _getterAxes(self, axesString, gcsFunction, valueArrayClass):
        nCh= len(axesString.split())
        value= valueArrayClass([0]* nCh)
        gcsFunction.argtypes= [c_int, c_char_p, valueArrayClass]
        self._convertErrorToException(
            gcsFunction(self._id, axesString.encode(), value))
        return value.toNumpyArray()


    def _setterAxes(self, axesString, value, gcsFunction, valueArrayClass):
        nCh= len(axesString.split())
        valueArray= np.atleast_1d(value)
        assert nCh == len(valueArray)
        gcsFunction.argtypes= [c_int, c_char_p, valueArrayClass]
        self._convertErrorToException(
            gcsFunction(self._id, axesString.encode(),
                        valueArrayClass(valueArray)))


    def _getterReturnString(self, gcsFunction, bufSize):
        s=ctypes.create_string_buffer(b'\000', bufSize)
        self._convertErrorToException(
            gcsFunction(self._id, s, bufSize))
        return s.value.decode()


    def _isConnected(self):
        return self._toBool(self._lib.PI_IsConnected(self._id))


    def connectTCPIP(self, hostname, port=50000):
        ide= self._lib.PI_ConnectTCPIP(hostname.encode(), port)
        if ide == -1:
            errorId= self._lib.PI_GetError(ide)
            raise PIConnectionError("%s" % self._errAsString(errorId))
        self._id= ide
        self._hostname= hostname
        self._port= port


    def closeConnection(self):
        try:
            self._lib.PI_CloseConnection(self._id)
        except Exception:
            pass


    def _trickToCheckForSyntaxError(self):
        import time
        time.sleep(0.1)
        self._convertErrorToException(0, 1)


    def gcsCommand(self, commandAsString):
        self._lib.PI_GcsCommandset.argtypes= [c_int, c_char_p]
        self._lib.PI_GcsGetAnswer.argtypes= [c_int, c_char_p, c_int]
        self._convertErrorToException(
            self._lib.PI_GcsCommandset(self._id, commandAsString.encode()))
        self._trickToCheckForSyntaxError()
        retSize= c_int()
        res= ''
        self._convertErrorToException(
            self._lib.PI_GcsGetAnswerSize(self._id, ctypes.byref(retSize)))
        while retSize.value != 0:
            buf= ctypes.create_string_buffer(b'\000', retSize.value)
            self._convertErrorToException(
                self._lib.PI_GcsGetAnswer(self._id, buf, retSize.value))
            res+= buf.value.decode()
            self._convertErrorToException(
                self._lib.PI_GcsGetAnswerSize(self._id, ctypes.byref(retSize)))
        return res


    def getVersion(self):
        bufSize= 256
        return self._getterReturnString(self._lib.PI_qVER, bufSize)


    def getAxesIdentifiers(self):
        return self._getterReturnString(self._lib.PI_qSAI, 256).split()


    def getNumberOfInputSignalChannels(self):
        nChannels= c_int()
        self._convertErrorToException(
            self._lib.PI_qTSC(self._id, ctypes.byref(nChannels)))
        return nChannels.value


    def getNumberOfOutputSignalChannels(self):
        nChannels= c_int()
        self._convertErrorToException(
            self._lib.PI_qTPC(self._id, ctypes.byref(nChannels)))
        return nChannels.value


    def getServoControlMode(self, axesString):
        return self._getterAxes(
            axesString, self._lib.PI_qSVO, CIntArray).astype(np.bool)


    def setServoControlMode(self, axesString, controlMode):
        self._setterAxes(
            axesString,
            np.atleast_1d(controlMode).astype('int'),
            self._lib.PI_SVO,
            CIntArray)


    def getControlMode(self, channels):
        return self._getterChannels(channels, self._lib.PI_qONL, CIntArray)


    def setControlMode(self, channels, controlMode):
        self._setterChannels(
            channels, controlMode, self._lib.PI_ONL, CIntArray)


    def enableControlMode(self, channels):
        enable= [CHANNEL_ONLINE]* len(channels)
        self.setControlMode(channels, enable)


    def disableControlMode(self, channels):
        disable= [CHANNEL_OFFLINE]* len(channels)
        self.setControlMode(channels, disable)


    def echo(self, message):
        cMsg= ctypes.create_string_buffer(message)
        cRet= ctypes.create_string_buffer(0, ctypes.sizeof(cMsg))
        self._convertErrorToException(
            self._lib.PI_qECO(self._id, cMsg, cRet))
        return cRet.value.decode()


    def getLowerVoltageLimit(self, channels):
        return self._getterChannels(channels, self._lib.PI_qVMI, CDoubleArray)


    def setLowerVoltageLimit(self, channels, lowerVoltage):
        self._setterChannels(
            channels, lowerVoltage, self._lib.PI_VMI, CDoubleArray)


    def getUpperVoltageLimit(self, channels):
        return self._getterChannels(channels, self._lib.PI_qVMA, CDoubleArray)


    def setUpperVoltageLimit(self, channels, upperVoltage):
        self._setterChannels(
            channels, upperVoltage, self._lib.PI_VMA, CDoubleArray)


    def getPosition(self, axesString):
        return self._getterAxes(axesString, self._lib.PI_qPOS, CDoubleArray)


    def getVoltages(self, channels):
        return self._getterChannels(channels, self._lib.PI_qVOL, CDoubleArray)


    def getOpenLoopAxisValue(self, axesString):
        return self._getterAxes(
            axesString, self._lib.PI_qSVA, CDoubleArray)


    def setOpenLoopAxisValue(self, axesString, amplitudeInVolt):
        self._setterAxes(
            axesString, amplitudeInVolt, self._lib.PI_SVA, CDoubleArray)


    def setRelativeOpenLoopAxisValue(self, axesString, offsetInVolt):
        self._setterAxes(
            axesString, offsetInVolt, self._lib.PI_SVR, CDoubleArray)


    def getTargetPosition(self, axesString):
        return self._getterAxes(
            axesString, self._lib.PI_qMOV, CDoubleArray)


    def setTargetPosition(self, axesString, position):
        self._setterAxes(
            axesString, position, self._lib.PI_MOV, CDoubleArray)


    def setTargetRelativeToCurrentPosition(self, axesString, offset):
        self._setterAxes(
            axesString, offset, self._lib.PI_MVR, CDoubleArray)


    def getVolatileMemoryParameters(self, itemId, parameterId):
        self._lib.PI_qSPA.argtypes= [c_int,
                                     c_char_p,
                                     CUnsignedIntArray,
                                     CDoubleArray,
                                     c_char_p,
                                     c_int]
        retValue= CDoubleArray([0.])
        bufSize= 256
        retString= ctypes.create_string_buffer(b'\000', bufSize)

        self._convertErrorToException(
            self._lib.PI_qSPA(
                self._id,
                str(itemId).encode(),
                CUnsignedIntArray([parameterId]),
                retValue,
                retString,
                bufSize))
        return retValue.toNumpyArray()


    def getAllDataRecorderOptions(self):
        bufSize= 1024
        return self._getterReturnString(self._lib.PI_qHDR, bufSize).split('\n')


    def getNumberOfRecorderTables(self):
        return int(self.getVolatileMemoryParameters(1, 0x16000300))


    def setDataRecorderConfiguration(self, dataRecorderConfiguration):
        assert isinstance(dataRecorderConfiguration,
                          DataRecorderConfiguration), \
            "argument must be of type DataRecorderConfiguration"

        self._lib.PI_DRC.argtypes= [c_int, CIntArray, c_char_p, CIntArray]

        for tableId in dataRecorderConfiguration.getTableIds():
            source= dataRecorderConfiguration.getRecordSource(tableId)
            option= dataRecorderConfiguration.getRecordOption(tableId)
            self._convertErrorToException(
                self._lib.PI_DRC(self._id,
                                 CIntArray([tableId]),
                                 source.encode(),
                                 CIntArray([option])))


    def getDataRecorderConfiguration(self):
        nRecorders= self.getNumberOfRecorderTables()
        sourceBufSize= 256
        source= ctypes.create_string_buffer(b'\000', sourceBufSize)
        option= CIntArray(np.zeros(nRecorders, dtype=np.int32))
        table=CIntArray(np.arange(1, nRecorders + 1))

        self._lib.PI_qDRC.argtypes= [c_int, CIntArray, c_char_p,
                                     CIntArray, c_int, c_int]

        self._convertErrorToException(
            self._lib.PI_qDRC(self._id, table, source,
                              option, sourceBufSize, nRecorders))

        sources= [x.strip() for x in source.value.decode().split('\n')]
        cfg= DataRecorderConfiguration()
        for i in range(nRecorders):
            cfg.setTable(table.toNumpyArray()[i],
                         sources[i],
                         option.toNumpyArray()[i])
        return cfg


    def getRecordedDataValues(self, howManyPoints, startFromPoint=1):
        nRecorders= self.getNumberOfRecorderTables()
        value= CDoubleArray(np.zeros(howManyPoints))
        retBuf= np.zeros((nRecorders, howManyPoints))
        self._lib.PI_qDRR_SYNC.argtypes=[c_int, c_int, c_int,
                                         c_int, CDoubleArray]

        for i in range(nRecorders):
            self._convertErrorToException(
                self._lib.PI_qDRR_SYNC(
                    self._id, i + 1, startFromPoint,
                    len(value.toNumpyArray()), value))
            retBuf[i, :]= value.toNumpyArray()

        return retBuf


    def startRecordingInSyncWithWaveGenerator(self):
        self._convertErrorToException(
            self._lib.PI_WGR(self._id))


    def getNumberOfWaveGenerators(self):
        return 3


    def getWaveGeneratorStartStopMode(self):
        nWaveGenerators= self.getNumberOfWaveGenerators()
        self._lib.PI_qWGO.argtypes= [c_int, CIntArray, CIntArray, c_int]
        wgIds= CIntArray(np.arange(1, nWaveGenerators+ 1))
        values= CIntArray(np.zeros(nWaveGenerators))
        self._convertErrorToException(
            self._lib.PI_qWGO(self._id, wgIds, values, nWaveGenerators))
        return values.toNumpyArray()


    def setWaveGeneratorStartStopMode(self, startModeArray):
        nWaveGenerators= self.getNumberOfWaveGenerators()
        self._lib.PI_WGO.argtypes= [c_int, CIntArray, CIntArray, c_int]
        wgIds= CIntArray(np.arange(1, nWaveGenerators+ 1))
        values= CIntArray(startModeArray)
        self._convertErrorToException(
            self._lib.PI_WGO(self._id, wgIds, values, nWaveGenerators))


    def clearWaveTableData(self, waveTableIdsArray):
        self._lib.PI_WCL.argtypes= [c_int, CIntArray, c_int]
        table= CIntArray(waveTableIdsArray)
        self._convertErrorToException(
            self._lib.PI_WCL(self._id, table, len(waveTableIdsArray)))


    def getConnectionOfWaveTableToWaveGenerator(self, waveGeneratorsArray):
        self._lib.PI_qWSL.argtypes= [c_int, CIntArray, CIntArray, c_int]
        nItems= len(waveGeneratorsArray)
        waveGenerators= CIntArray(waveGeneratorsArray)
        waveTable= CIntArray(np.zeros(nItems))
        self._convertErrorToException(
            self._lib.PI_qWSL(self._id,
                              waveGenerators,
                              waveTable,
                              nItems))
        return waveTable.toNumpyArray()


    def setConnectionOfWaveTableToWaveGenerator(self,
                                                waveGeneratorsArray,
                                                waveTableIdsArray):
        assert len(waveGeneratorsArray) == len(waveTableIdsArray)
        self._lib.PI_WSL.argtypes= [c_int, CIntArray, CIntArray, c_int]
        waveGenerators= CIntArray(waveGeneratorsArray)
        waveTable= CIntArray(waveTableIdsArray)
        self._convertErrorToException(
            self._lib.PI_WSL(self._id,
                             waveGenerators,
                             waveTable,
                             len(waveTableIdsArray)))




    def setSinusoidalWaveform(self,
                              waveTableId,
                              append,
                              lengthInPoints,
                              amplitudeOfTheSineCurve,
                              offsetOfTheSineCurve,
                              wavelengthOfTheSineCurveInPoints,
                              startPoint,
                              curveCenterPoint):
        '''
        See description of PI_WAV_SIN_P in PI GCS 2.0 DLL doc
        '''
        assert append in WaveformGenerator.ALL

        self._lib.PI_WAV_SIN_P.argtypes= [c_int, c_int, c_int, c_int, c_int,
                                          c_int, c_double, c_double, c_int]
        self._convertErrorToException(
            self._lib.PI_WAV_SIN_P(self._id,
                                   waveTableId,
                                   int(startPoint),
                                   int(wavelengthOfTheSineCurveInPoints),
                                   append,
                                   int(curveCenterPoint),
                                   amplitudeOfTheSineCurve,
                                   offsetOfTheSineCurve,
                                   int(lengthInPoints)))



    def setRecordTableRate(self, recordTableRateInServoLoopCycles=1):
        self._lib.PI_RTR.argtypes= [c_int, c_int]
        self._convertErrorToException(
            self._lib.PI_RTR(self._id, int(recordTableRateInServoLoopCycles)))


    def getRecordTableRate(self):
        rtr= c_int()
        self._convertErrorToException(
            self._lib.PI_qRTR(self._id, ctypes.byref(rtr)))
        return rtr


    def getServoUpdateTimeInSeconds(self):
        return self.getVolatileMemoryParameters(1, 0x0E000200)




    def setWaveGeneratorTableRate(self,
                                  waveGeneratorTableRateInServoLoopCycles):
        self._lib.PI_WTR.argtypes= [c_int, CIntArray, CIntArray,
                                    CIntArray, c_int]
        nWaveGenerators= self.getNumberOfWaveGenerators()
        wgIds= CIntArray(np.arange(1, nWaveGenerators+ 1))
        tableRate= CIntArray(waveGeneratorTableRateInServoLoopCycles)
        interpolation= CIntArray(np.zeros(nWaveGenerators))

        self._convertErrorToException(
            self._lib.PI_WTR(self._id,
                             wgIds,
                             tableRate,
                             interpolation,
                             nWaveGenerators))


    def getWaveGeneratorTableRate(self):
        self._lib.PI_qWTR.argtypes= [c_int, CIntArray, CIntArray,
                                     CIntArray, c_int]

        nWaveGenerators= self.getNumberOfWaveGenerators()
        wgIds= CIntArray(np.arange(1, nWaveGenerators+ 1))
        tableRate= CIntArray(np.zeros(nWaveGenerators))
        interpolation= CIntArray(np.zeros(nWaveGenerators))
        self._convertErrorToException(
            self._lib.PI_qWTR(self._id,
                              wgIds,
                              tableRate,
                              interpolation,
                              nWaveGenerators))
        return tableRate.toNumpyArray()


    def getOverflowState(self, axesString):
        return self._getterAxes(
            axesString, self._lib.PI_qOVF, CIntArray).astype(np.bool)


    def setDataRecoderTriggerSource(self, source, value):
        pass


    def getDataRecorderTriggerSource(self):
        pass



    def startStepAndResponseMeasurement(self, axisString, amplitude):
        self._lib.PI_STE.argtypes= [c_int, c_char_p, c_double]

        self._convertErrorToException(
            self._lib.PI_STE(self._id,
                             axisString.encode(),
                             amplitude))


    def startImpulseAndResponseMeasurement(self, axisString, amplitude):
        self._lib.PI_IMP.argtypes= [c_int, c_char_p, c_double]

        self._convertErrorToException(
            self._lib.PI_IMP(self._id,
                             axisString.encode(),
                             amplitude))

